local addonName, vars = ...
local L = vars.L
GarajalAnnounce = LibStub("AceAddon-3.0"):NewAddon(addonName)
local addon = GarajalAnnounce 

local defaults = {
  profile = {
    debug = true, -- for addon debugging
    dryrun = true,
    whisper = true,
    raidwarning = true,
    raidicons = true,
    minimap = {
        hide = false,
    },
    players = {},
  }
}

local playertemp = {
  ["Spernatral"] = { priority=1 },
  ["Fangv"] = { priority=3 },
  ["Yuzuriha"] = { priority=3 },

  ["Howdy"] = { priority=1 },
  ["Nutmeg"] = { priority=3 },
}
local players
local defaultpriority = 2

local spellVoodoo =  122151  -- voodoo doll, ineligble for totem
local spellCrossed = 116161  -- Crossed Over, in spirit realm up to 30 sec
local spellBuffed =  117549  -- Spiritual innervation, 30 sec dps/mana buff
local spellFrail =   117723  -- Frail Soul (heroic) cannot enter for 30 sec
local spellTotem =   116174  -- boss summoned a new totem

local spellNameCrossed = GetSpellInfo(spellCrossed)
local spellNameBuffed = GetSpellInfo(spellBuffed)

local settings = defaults.profile
local optionsFrame
local charName
local revision = tonumber(("$Revision: 3 $"):match("%d+"))
local minimapIcon = LibStub("LibDBIcon-1.0")
local LDB, LDBo

local function chatMsg(msg)
     DEFAULT_CHAT_FRAME:AddMessage(addonName..": "..msg)
end
local function debug(msg)
  if addon.db.profile.debug then
     chatMsg(msg)
  end
end

local function UnitToName(unit)
  return GetUnitName(unit, true)
end

function addon:GetDiff()
  local instname, insttype, diff, diffname, maxPlayers, playerDifficulty, isDynamicInstance = GetInstanceInfo()
  addon.raidsize = maxPlayers
  if HasLFGRestrictions() or IsInLFGDungeon() then
     addon.raiddiff = "lfr"
  elseif diff == 5 or diff == 6 then
     addon.raiddiff = "heroic"
  else
     addon.raiddiff = "normal"
  end
end

function addon:canAnnounce()
  return UnitIsGroupLeader("player") or UnitIsGroupAssistant("player")
end

function addon:InZone()
  local oldmap = GetCurrentMapAreaID()
  local oldlvl = GetCurrentMapDungeonLevel()
  SetMapToCurrentZone()
  local map = GetCurrentMapAreaID()
  local lvl = GetCurrentMapDungeonLevel()
  local inzone = (map == 896 and lvl == 2)
  local unitX, unitY = GetPlayerMapPosition("player")
  if not unitX then unitX,unitY = 0,0 end
  unitX, unitY = unitX*100, unitY*100
  SetMapByID(oldmap)
  if lvl and lvl > 0 then
    SetDungeonMapLevel(oldlvl)
  end
  return inzone, unitX, unitY
end

local combat_events = {
 "UNIT_SPELLCAST_SUCCEEDED",
 "PLAYER_REGEN_DISABLED",
 "UNIT_AURA",
 -- "COMBAT_LOG_EVENT_UNFILTERED"
}

function addon:CheckEnable()
  if not addon.init then return end
  --debug("CheckEnable")
  if addon:InZone() and not addon:canAnnounce() and not addon.offwarn then
    chatMsg(L["You must have raid lead or assist to make annoucements."])
    addon.offwarn = true
  end
  if addon:InZone() and (addon:canAnnounce() or addon.db.profile.debug) then
    addon:ScanRaid()
    if addon.enabled then return end
    addon.enabled = true
    addon.offwarn = false
    chatMsg(L["Addon enabled. Type '/ga config' for options."])
    for _,evt in pairs(combat_events) do 
      addon:RegisterEvent(evt)
    end
  else -- wrong zone
    if not addon.enabled then return end
    chatMsg(L["Addon disabled."])
    addon.enabled = false
    for _,evt in pairs(combat_events) do 
      addon:UnregisterEvent(evt)
    end
  end

end

function addon:ScanRaid(reset)
  addon:GetDiff()
  local cnt = 0
  local unassigned = 0
  for i=1,40 do
    local unit = "raid"..i
    if UnitExists(unit) then
      cnt = cnt + 1
      local name = UnitToName(unit)
      local info = players[name] or {}
      players[name] = info
      info.unit = unit
      info.name = name
      info.role = UnitGroupRolesAssigned(unit)
      if info.role == "NONE" then
        unassigned = unassigned + 1
      end
      info.priority = (reset and defaultpriority) or info.priority or defaultpriority
      if info.priority < 1 or info.priority > 3 then
         info.priority = defaultpriority
      end
      info.order = (reset and cnt) or info.order or cnt
      info.lastaction = 0
      debug(name..":"..info.role..":"..info.priority..":"..info.order)
    end
  end
  if unassigned > 0 then
    chatMsg(string.format(L["WARNING: %d players do not have roles assigned and will be ignored."],unassigned))
  end
  debug("Processed "..cnt.." players, "..unassigned.." unassigned roles")
end

local function playerSort(i1, i2)
  if (i1.delay ~= i2.delay) then 
    return i1.delay < i2.delay
  elseif (i1.role == "DAMAGER" and i1.priority ~= i2.priority) then
    return i1.priority < i2.priority
  elseif (i1.role == "HEALER" and i1.mana ~= i2.mana) then
    return i1.mana < i2.mana
  elseif (i1.lastaction ~= i2.lastaction) then
    return i1.lastaction < i2.lastaction
  else
    return i1.order <= i2.order
  end
end

local healerElig = {}
local dpsElig = {}
local playersChosen = {}
function addon:handleTotem()
  local now = GetTime()
  addon.totemcount = (addon.totemcount or 0) + 1
  debug("handleTotem() "..addon.totemcount);
  debug(GetRealZoneText()..":"..GetSubZoneText()..":"..GetZoneText()..":"..GetMinimapZoneText())
  wipe(healerElig)
  wipe(dpsElig)
  for i=1,40 do
    local unit = "raid"..i
    local zone, online, isDead = select(7,GetRaidRosterInfo(i))
    if UnitExists(unit) and (settings.dryrun or 
       (online and not isDead and zone == addon.zone and
        not UnitIsDeadOrGhost(unit) and UnitIsConnected(unit))) then
       local name = UnitToName(unit)
       local info = players[name]
       local list
       if info and info.role == "DAMAGER" then 
         list = dpsElig
       elseif info and info.role == "HEALER" then 
         list = healerElig
         info.mana = UnitPower(unit, SPELL_POWER_MANA)/UnitPowerMax(unit, SPELL_POWER_MANA)
       end
       local elig = list
       if elig then
         info.delay = 0
         for j=1,40 do
           local spellname, _,_,_,_,_, expires, _,_,_, spellid = UnitAura(unit, j)
           if not spellname then break end
           if spellid == spellVoodoo then
             elig = false; 
             break
           end
           if (spellid == spellCrossed and addon.raiddiff == "heroic") then
             elig = false; 
             break             
           elseif (spellid == spellFrail or spellid == spellBuffed or spellid == spellCrossed) then
             local remain = (expires - now)
             info.delay = math.max(info.delay, remain)     
           end
         end
       end
       if elig then
         debug("Elig "..info.role..": "..name.."  zone="..zone)
         table.insert(list,info)
       end
    end -- unit alive
  end  -- raid members
  table.sort(healerElig, playerSort)
  table.sort(dpsElig, playerSort)
  local healercnt = 1
  local dpscnt = 2
  local rtmod = 6
  if addon.raidsize > 10 then
    dpscnt = 3
    rtmod = 8
  end
  wipe(playersChosen)
  for i=1,healercnt do
    if healerElig[i] then
      table.insert(playersChosen, healerElig[i])
    else
      chatMsg(L["WARNING: Not enough eligible healers!"])
    end
  end
  for i=1,dpscnt do
    if dpsElig[i] then
      table.insert(playersChosen, dpsElig[i])
    else
      chatMsg(L["WARNING: Not enough eligible dps!"])
    end
  end
  local msg = L["Spirit Totem"].." " .. addon.totemcount..":"
  local delay = 0
  addon.rt = addon.rt or 0
  for idx,info in ipairs(playersChosen) do
    delay = math.max(delay, info.delay)
    info.lastaction = now
    local rt = addon.rt + 1
    addon.rt = (addon.rt + 1) % rtmod
    msg = msg .. " "
    if settings.raidicons then
      if addon:canAnnounce() and not addon.suppress then
        SetRaidTarget(info.unit, rt)
      end
      msg = msg .. "{rt"..rt.."}"     
    end
    msg = msg .. info.name
    if addon:canAnnounce() and settings.whisper and not addon.suppress then
      SendChatMessage(L["Run to the totem!!!"], "WHISPER", nil, info.name)
    end
  end
  if delay > 1000 then
    msg = " ("..L["DELAY"]..SecondsToTime(delay/1000)..")"
  end
  debug(msg)
  if addon:canAnnounce() and settings.raidwarning and not addon.suppress then
    SendChatMessage(msg, "RAID_WARNING")
  end
end

function addon:PLAYER_REGEN_DISABLED() -- entering combat
  addon.totemcount = 0
  addon.zone = GetRealZoneText()
end

function addon:UNIT_AURA(unit) -- detect actual cross-overs
  local expires = select(7,UnitAura(unit, spellNameCrossed))
  if expires then -- has crossed aura
    local now = GetTime()
    if expires > now + 25000 then -- just crossed
       local name = UnitToName(unit)
       local info = players[name]
       if info then
          info.lastaction = now
       end
    end
  end
  if settings.raidicons and UnitAura(unit, spellNameBuffed) and GetRaidTargetIndex(unit) then
     SetRaidTarget(unit, 0) -- got buffed, remove icon
  end
end

function addon:COMBAT_LOG_EVENT_UNFILTERED(...)
  local event,_,_,_,_,_,_,_,_,_,id = ...
  if event == "SPELL_CAST_SUCCESS" and id == spellTotem then
    addon:handleTotem()
  end
end

function addon:UNIT_SPELLCAST_SUCCEEDED(event, unitid, spellName, _, _, spellID)
  if spellID == spellTotem then
    addon:handleTotem()
  end 
end

function addon:myOptions() 
return {
  type = "group",
  set = function(info,val)
          local s = settings ; for i = 2,#info-1 do s = s[info[i]] end
          s[info[#info]] = val; debug(info[#info].." set to: "..tostring(val))
          addon:Update()
        end,
  get = function(info)
          local s = settings ; for i = 2,#info-1 do s = s[info[i]] end
          return s[info[#info]] end,
  args = {
   general = {
    type = "group",
    inline = true,
    name = L["General"],
    args = {
      debug = {
        name = L["Debug"],
        desc = L["Toggle debugging output"],
        type = "toggle",
        guiHidden = true,
      },
      config = {
        name = L["Config"],
        desc = L["Open the configuration GUI"],
        type = "execute",
        guiHidden = true,
        func = function() addon:Config() end,
      },
      minimap = {
        order = 15,
        name = L["Minimap Icon"],
        desc = L["Display minimap icon"],
        type = "toggle",
        set = function(info,val)
          settings.minimap.hide = not val
          addon:Update()
	end,
        get = function() return not settings.minimap.hide end,
      },
      aheader = {
        name = L["Announcement Options"],
        type = "header",
        cmdHidden = true,
        order = 99,
      },
      raidwarning = {
        order = 100,
        name = L["Raid Warning"],
        desc = L["Send a raid warning listing selected players"],
        type = "toggle",
      },
      whisper = {
        order = 110,
        name = L["Whisper"],
        desc = L["Whisper selected players"],
        type = "toggle",
      },
      raidicons = {
        order = 120,
        name = L["Raid Icons"],
        desc = L["Set raid icons on selected players"],
        type = "toggle",
      },
      rheader = {
        name = L["Raid"],
        type = "header",
        cmdHidden = true,
        order = 199,
      },
      rolecheck = {
        order = 200,
        name = L["Role Check"],
        desc = L["Start a Raid Role Check"],
        type = "execute",
	disabled = function() return not addon:canAnnounce() end,
        func = function() InitiateRolePoll() end,
      },
     },
    },
  }
} 
end

local function table_clone(t)
  if not t then return nil
  elseif type(t) == "table" then
    local res = {}
    for k,v in pairs(t) do
      res[table_clone(k)] = table_clone(v)
    end
    return res
  else
    return t
  end
end

function addon:RefreshConfig()
  -- things to do after load or settings are reset
  debug("RefreshConfig")
  settings = addon.db.profile
  addon.settings = settings
  charName = UnitName("player")
  for k,v in pairs(defaults.profile) do
     if settings[k] == nil then
       settings[k] = table_clone(v)
     end
  end
  players = settings.players
  players = playertemp
  addon.players = players
  settings.loaded = true
  addon:Update()
end

function addon:Update()
  -- things to do when settings change
  if LDBo then
    if settings.minimap.hide then
      minimapIcon:Hide(addonName)
    else
      minimapIcon:Show(addonName)
    end
  end
  addon:CheckEnable()
end

function addon:OnInitialize()
  addon.db = LibStub("AceDB-3.0"):New("GarajalAnnounceDB", defaults)
  addon:RefreshConfig()
  local options = addon:myOptions()
  LibStub("AceEvent-3.0"):Embed(addon)
  LibStub("AceConfigRegistry-3.0"):ValidateOptionsTable(options, addonName)
  LibStub("AceConfig-3.0"):RegisterOptionsTable(addonName, options, {"garajalannounce", "ga"})
  optionsFrame = LibStub("AceConfigDialog-3.0"):AddToBlizOptions(addonName, addonName, nil, "general")
  optionsFrame.default = function()
       for k,v in pairs(defaults.profile) do settings[k] = table_clone(v) end
       addon:RefreshConfig()
       if InterfaceOptionsFrame:IsShown() then
         addon:Config(); addon:Config()
       end
  end
  options.args.profiles = LibStub("AceDBOptions-3.0"):GetOptionsTable(addon.db)
  LibStub("AceConfigDialog-3.0"):AddToBlizOptions(addonName, L["Profiles"], addonName, "profiles")

  debug("OnInitialize")

  self.db.RegisterCallback(self, "OnProfileChanged", "RefreshConfig")
  self.db.RegisterCallback(self, "OnProfileCopied", "RefreshConfig")
  self.db.RegisterCallback(self, "OnProfileReset", "RefreshConfig")
  self.db.RegisterCallback(self, "OnDatabaseReset", "RefreshConfig")
  addon.init = true
end

function addon:Config()
  if optionsFrame then
    if ( InterfaceOptionsFrame:IsShown() ) then
      InterfaceOptionsFrame:Hide();
    else
      InterfaceOptionsFrame_OpenToCategory(optionsFrame)
    end
  end
end

function addon:RunCheck()
  debug("RunCheck")

end

function addon:OnEnable()
  debug("OnEnable")
  self:RegisterEvent("GROUP_ROSTER_UPDATE", "CheckEnable")
  self:RegisterEvent("ZONE_CHANGED","CheckEnable")
  self:RegisterEvent("ZONE_CHANGED_INDOORS","CheckEnable")
  self:RegisterEvent("ZONE_CHANGED_NEW_AREA","CheckEnable")
 
  if LDB then
    return
  end
  if AceLibrary and AceLibrary:HasInstance("LibDataBroker-1.1") then
    LDB = AceLibrary("LibDataBroker-1.1")
  elseif LibStub then
    LDB = LibStub:GetLibrary("LibDataBroker-1.1",true)
  end
  if LDB then
    LDBo = LDB:NewDataObject(addonName, {
        type = "launcher",
        label = addonName,
        icon = "Interface\\Icons\\spell_shadow_mindtwisting",
        OnClick = function(self, button)
                if button == "RightButton" then
                        addon:Config()
                else
                        addon:RunCheck()
                end
        end,
        OnTooltipShow = function(tooltip)
                if tooltip and tooltip.AddLine then
                        tooltip:SetText(addonName)
                        tooltip:AddLine("|cffff8040"..L["Left Click"].."|r "..L["to toggle status window"])
                        tooltip:AddLine("|cffff8040"..L["Right Click"].."|r "..L["for options"])
                        tooltip:Show()
                end
        end,
     })
  end 

  if LDBo then
    minimapIcon:Register(addonName, LDBo, settings.minimap)
  end
  addon:Update()
end



